/* Written by Peter L Higginson, Sophos PLC
 * 16 March 2004
 *
 * Simple program to dump out the internal state of the Dazuko Kernel Module
 *
 */
 
/* NOTE - THIS IS WORK IN PROGRESS, I add extra structures as I need them
 *
 * It expects to find the Dazuko module at /opt/sophos/sav/dazuko/dazuko.o
 * you will probably need to change this
 *
 * It was written mainly to look at hangs - the current code loops dumping the state
 * - look around line 264 to change this
 * 
 * It has the Linux (2.4) structures included directly, so will only work on that platform.
 * I did this to avoid any problems from including structures that conflicted with user
 * mode structures. I also needed to find them to understand what was being dumped.
 * Hint - I used a -E compile of dazuko_xp.c (and an editor) to get these.
 */



#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

extern int errno;

/* slot and other structures */

#define NUM_SLOTS 25

typedef struct { volatile int counter; } atomic_t;

struct list_head {
        struct list_head *next, *prev;
};
typedef struct list_head list_t;

typedef struct {
        volatile unsigned int lock;
} spinlock_t;

struct __wait_queue_head {
        spinlock_t lock;
        struct list_head task_list;
};
typedef struct __wait_queue_head wait_queue_head_t;

struct semaphore {
        atomic_t count;
        int sleepers;
        wait_queue_head_t wait;
};

struct xp_mutex
{
        struct semaphore mutex;
};

struct xp_atomic
{
        atomic_t        atomic;
};

struct daemon_id
{
    int pid;
    int unique;
    void* f_ref;
};

struct event_properties
{
    int flags;
    char    set_flags;
    int mode;
    char    set_mode;
    int uid;
    char    set_uid;
    int pid;
    char    set_pid;
};

struct file_properties
{
    unsigned long   size;
    char        set_size;
    int     uid;
    char        set_uid;
    int     gid;
    char        set_gid;
    int     mode;
    char        set_mode;
    int     device_type;
    char        set_device_type;
};

struct slot
{
    /* A representation of a daemon. It holds
     * all information about the daemon, the
     * file that is scanned, and the state of
     * the scanning process. */

    int         id;
    struct daemon_id    did;        /* identifier for our daemon */
    int         write_mode;
    int         state;
    int         response;
    int         event;
    int         filenamelength; /* not including terminator */
    char            *filename;
    struct event_properties event_p;
    struct file_properties  file_p;
    struct xp_mutex     mutex;
};

struct slot_list
{
    struct xp_atomic    use_count;
    struct slot     slots[NUM_SLOTS];
    char            reg_name[100];    /* this MUST be at the end of the struct - guess 100 here */
};

struct slot_list slot_data;

struct xp_queue
{
        wait_queue_head_t queue;
};


/* end of slot and other structures */

void list_queue(int kmem, int next, int prev, int curpos)
{
    int kread[20];
    int max = 29;                /* race conditions may break links - we have no locks - max 28 links */
    int curr = curpos - 8;       /* assumes list head is last two locations read */

    if (next == curr) return;
    
    /* I tried reading the current position but it failed */

    while (next != (curpos - 8) && --max > 0) {
        if (lseek(kmem, next, SEEK_SET) == -1) {
            printf("Seek Failed in list_queue for %x\n", next);
            continue;
        }
        if (read(kmem, &kread[0], 24) != 24) {
          printf("Read Failed in list_queue for %x\n", next);
          continue;
        }
        printf("        List entry at %x is %x %x %x %x %x %x\n", next, kread[0], kread[1], kread[2], kread[3]);
        if (curr != kread[1]) {
            printf("          Link break in list_queue curr = %x\n",curr);
            break;
        }
        curr = next;
        next = kread[0];
    }

    if (lseek(kmem, curpos, SEEK_SET) == -1) {
      printf("Seek Reset failed in list_queue for %x\n", curpos);
      return;
    }
}


int main()
{
    size_t n = 0;
    size_t c;
    char *buf = NULL;
    int value, dazg = -1, slotl = -1, seekpos;
    int rodata = 0, vdata = 0;
    char* list[200];
    int listCount = 0;
    int kmem, count, scnt;
    char x, var[100], path[1024];
    int kread[20];
    int slot[5] = {0,0,0,0,0};
    int sl, runcnt;
    
    system("/sbin/ksyms | grep dazuko > tmp_info.txt");
    system("nm /opt/sophos/sav/dazuko/dazuko.o >> tmp_info.txt");
    
    FILE *info = fopen("tmp_info.txt","r");
    if (info == NULL) {
        printf("No info file\n");
        return 0;
    }
  
    while ((c = getline(&buf,&n,info)) != -1) {
        if (c <= 1) continue;
        if (strstr(buf,"[dazuko]")) {
            value = 0;
            sscanf(buf,"%x ",&value);
            if (value) {
                if (strstr(buf,"rodata")) rodata = value;
                else if (strstr(buf,"data")) vdata = value;
            }
            continue;
        }
        /* here with normal variables */
        if (buf[9] == 'b' || buf[9] == 'B' || buf[9] == 'd' || buf[9] == 'D' || buf[9] == 'r' || buf[9] == 'R') {
            if (listCount >= 199) {
                printf("List too small\n");
                return 0;
            }
            list[listCount++] = buf;
            buf = NULL;
            n = 0;
        }
    }
    fclose(info);
    free(buf);
    buf = NULL;
    
    if (vdata == 0) {
        printf("Failed to find data address (check Dazuko running)\n");
        return 0;
    }
    
    kmem = open("/dev/kmem",O_RDONLY);
    if (kmem == -1) {
        printf("Cannot open /dev/kmem\n");
        return 0;
    }
    
    /* find where the variable data is */
    var[99] = '\0';
    for (count = 0; count < listCount; ++count) {
        if ((c = sscanf(list[count],"%x %c %99s\n",&value, &x, &var)) != 3) {
            printf("Fail c=%d, scan line %s",c,list[count]);
            continue;
        }
        if (strcmp(var,"dazuko_global") == 0) dazg = value;
        if (strcmp(var,"slot_lists") == 0) slotl = value;
    }

    if (dazg == -1 || slotl == -1) {
        printf("Failed to find dazuko_global or slot_lists in symbol table\n");
        return 0;
    }        
        
    if (lseek(kmem, vdata+dazg, SEEK_SET) == -1) {
        printf("Seek Failed for dazuko_global\n");
        return 0;
    }
    if (read(kmem, &dazg, 4) != 4) {
        printf("Read Failed for dazuko_global\n");
        return 0;
    }
    dazg -= slotl;
    
    printf("Data at 0x%x, ROdata at 0x%x, Uninitdata at 0x%x\n",vdata, rodata, dazg);
    
    /*  for (runcnt = 0; runcnt < 5; ++runcnt) { */
    while(1) {

        var[99] = '\0';
        for (count = 0; count < listCount; ++count) {
            if ((c = sscanf(list[count],"%x %c %99s\n",&value, &x, &var)) != 3) {
                printf("Fail c=%d, scan line %s",c,list[count]);
                continue;
            }
            /* printf("symbol %s at 0x%x type %c\n", var, value, x); */
            if (x == 'd' || x == 'D') {
                if (lseek(kmem, vdata+value, SEEK_SET) == -1) {
                    printf("Seek Failed for %s\n", var);
                    continue;
                }
                seekpos = vdata + value;
                
                /* default cases, ints and chars */
                if (read(kmem, &kread[0], 4) != 4) {
                    printf("Read Failed for %s\n", var);
                    continue;
                }
                /* catch the char variables */
                if (strcmp(var, "access_mask") == 0) kread[0] &= 0xff;
                printf("%s at 0x%x (init) is %d = 0x%x\n", var, value, kread[0], kread[0]);
            }
        
            if (x == 'b' || x == 'B') {
                if (lseek(kmem, dazg+value, SEEK_SET) == -1) {
                    printf("Seek Failed for %s\n", var);
                    continue;
                }
                seekpos = dazg + value;
            
                /* default case int */
                if (read(kmem, &kread[0], 4) != 4) {
                    printf("Read Failed for %s\n", var);
                    continue;
                }
                printf("%s at 0x%x (un-init) is %d = 0x%x\n", var, value, kread[0], kread[0]);
            }
            if (strcmp(var,"incl_paths") == 0 || strcmp(var,"excl_paths") == 0) {
                while ((value = kread[0]) != 0) {
                    if (lseek(kmem, value, SEEK_SET) == -1) {
                        printf("Contin Seek Failed for %s\n", var);
                        continue;
                    }
                    if (read(kmem, &kread[0], 8) != 8) {
                        printf("Contin Read Failed for %s\n", var);
                        continue;
                    }
                    if (kread[1] > 1022 || read(kmem, path, kread[1]) != kread[1]) {
                        printf("2nd Contin Read Failed for %s kread[1]=%d\n", var, kread[1]);
                        continue;
                    }
                    path[kread[1]] = '\0';
                    printf("    Path List len %d = %s, Next 0x%x\n", kread[1], path, kread[0]);
                }
            }
            if (strcmp(var,"slot_lists") == 0) {
                for (scnt = 0; scnt < 5; ++scnt) {   /* two loops so can just read */
                    if (read(kmem, &kread[1], 20) != 20) {
                        printf("Contin Read Failed for %s\n", var);
                        continue;
                    }
                    printf("   Slot %d Entry %08x %08x %08x %08x %08x %08x\n", scnt, kread[0],kread[1],kread[2],kread[3],kread[4],kread[5]);
                    slot[scnt] = kread[0];
                    list_queue(kmem, kread[4], kread[5], seekpos + (scnt * 24) + 24);
                    if (read(kmem, &kread[0], 4) != 4) {
                        printf("Contin2 Read Failed for %s\n", var);
                        continue;
                    }
                }
                for (scnt = 0; scnt < 5; ++scnt) {
                    if (slot[scnt]) {
                        if (lseek(kmem, slot[scnt], SEEK_SET) == -1) {
                            printf("Seek Failed for Slot %d\n", scnt);
                            continue;
                        }
    
                        if (read(kmem, &slot_data, sizeof(slot_data)) != sizeof(slot_data)) {
                            printf("Read Failed for slot %d\n", scnt);
                            continue;
                        }
    
                        slot_data.reg_name[99] = '\0';
                        printf("\n   Slot List %d, Use Count %d, Name %s\n", scnt, slot_data.use_count, &slot_data.reg_name[0]);
    
                        for (sl = 0; sl < NUM_SLOTS; ++sl) {
                            printf("      Slot %d, ID %d, DID %d-%d-%x\n", sl, slot_data.slots[sl].id, slot_data.slots[sl].did.pid,
                            slot_data.slots[sl].did.unique, slot_data.slots[sl].did.f_ref);
                            if (slot_data.slots[sl].did.pid) {
    
                                if (slot_data.slots[sl].filenamelength && slot_data.slots[sl].filenamelength < 1022) {
                                    if (lseek(kmem, (int)slot_data.slots[sl].filename, SEEK_SET) == -1) {
                                        printf("Seek Failed for Filename in Slot %d-%d\n", scnt, sl);
                                        continue;
                                    }
                                    if (read(kmem, path, slot_data.slots[sl].filenamelength) != slot_data.slots[sl].filenamelength) {
                                        printf("Read Failed for Filename in Slot %d-%d\n", scnt, sl);
                                        continue;
                                    }
                                    path[slot_data.slots[sl].filenamelength] = '\0';
                                } else path[0] = '\0';
    
                                printf("        Write Mode %d, State %d, Response %d, Event %d FilenameLength %d Filename %s\n",
                                    slot_data.slots[sl].write_mode, slot_data.slots[sl].state, slot_data.slots[sl].response,
                                    slot_data.slots[sl].event, slot_data.slots[sl].filenamelength, path);
                                printf("        Event (set) flags %x(%d), mode %x(%d), uid %d(%d), pid %d(%d)\n",
                                       slot_data.slots[sl].event_p.flags, slot_data.slots[sl].event_p.set_flags,
                                       slot_data.slots[sl].event_p.mode, slot_data.slots[sl].event_p.set_mode,
                                       slot_data.slots[sl].event_p.uid, slot_data.slots[sl].event_p.set_uid,
                                       slot_data.slots[sl].event_p.pid, slot_data.slots[sl].event_p.set_pid);
                                printf("        File (set) size %d(%d), uid %d(%d), gid %d(%d), mode %x(%d), device_type %x(%d)\n",
                                       slot_data.slots[sl].file_p.size, slot_data.slots[sl].file_p.set_size,
                                       slot_data.slots[sl].file_p.uid, slot_data.slots[sl].file_p.set_uid,
                                       slot_data.slots[sl].file_p.gid, slot_data.slots[sl].file_p.set_gid,
                                       slot_data.slots[sl].file_p.mode, slot_data.slots[sl].file_p.set_mode,
                                       slot_data.slots[sl].file_p.device_type, slot_data.slots[sl].file_p.set_device_type);
                                printf("        Mutex count %d, sleepers %d, lock %d, next %x, prev %x\n",
                                       slot_data.slots[sl].mutex.mutex.count.counter, slot_data.slots[sl].mutex.mutex.sleepers,
                                       slot_data.slots[sl].mutex.mutex.wait.lock.lock, slot_data.slots[sl].mutex.mutex.wait.task_list.next,
                                       slot_data.slots[sl].mutex.mutex.wait.task_list.prev);
                            }
                        }
                    }
                }
            }
            if (strcmp(var,"wait_kernel_waiting_for_free_slot") == 0 || strcmp(var,"wait_daemon_waiting_for_work") == 0 ||
                strcmp(var,"wait_kernel_waiting_while_daemon_works") == 0 || strcmp(var,"wait_daemon_waiting_for_free") == 0) {
                if (read(kmem, &kread[1], 8) != 8) {
                    printf("Contin Read Failed for %s\n", var);
                    continue;
                }
                printf("       Lock %d, Next %x, Prev %x\n", kread[0], kread[1], kread[2]);
                list_queue(kmem, kread[1], kread[2], seekpos + 12);
            }
        }

        sleep(10);
    }   /* loop here to get repeats of the snapshot */

    /* dump the data area in hex - to assist finding things */
    if (lseek(kmem, vdata, SEEK_SET) == -1) {
        printf("Seek Failed for vdata\n");
        return 0;
    }
    for (count = 0; count < 20; ++count) {
        if (read(kmem, &kread[0], 16) != 16) {
            printf("Read Failed for count %d\n", count);
            return 0;
        }
        printf("%08x - %08x %08x %08x %08x\n",vdata+count*16,kread[0],kread[1],kread[2],kread[3]);
    }
    close(kmem);
}
